import inspect

from orwynn.base.worker import Worker
from orwynn.bootscript.bootscript import Bootscript
from orwynn.bootscript.calltime import CallTime
from orwynn.bootscript.errors import (
    NoScriptsForCallTimeError,
    ScriptsAlreadyCalledError,
)
from orwynn.di.collecting.acceptordependencies import (
    collect_dependencies_for_acceptor,
)
from orwynn.di.container import DiContainer
from orwynn.di.provider import Provider

_IsCallPerformed = bool
# Represents bootscripts sorted by their call time and followed by
# a flag signified whether the call was already performed
_ScriptsState = tuple[list[Bootscript], _IsCallPerformed]
_CallState = dict[CallTime, _ScriptsState]


class BootscriptWorker(Worker):
    """
    Calls different defined bootscripts.
    """
    def __init__(self, bootscripts: list[Bootscript]) -> None:
        super().__init__()
        self.__bootscripts: list[Bootscript] = bootscripts
        self.__call_state: _CallState = self.__initialize_call_state()

    def __initialize_call_state(
        self
    ) -> _CallState:
        state: _CallState = {}

        for bs in self.__bootscripts:
            state.setdefault(bs.call_time, ([], False))
            state[bs.call_time][0].append(bs)

        return state

    async def call_by_time(
        self,
        call_time: CallTime,
        di_container: DiContainer
    ) -> None:
        try:
            scripts_state: _ScriptsState = \
                self.__call_state[call_time]
        except KeyError as err:
            raise NoScriptsForCallTimeError(
                f"cannot find any scripts for call time {call_time}"
            ) from err
        else:
            if scripts_state[1] is True:
                raise ScriptsAlreadyCalledError(
                    f"scripts for time {call_time} were already called"
                )
            else:
                for script in scripts_state[0]:
                    await self.__call_script(script, di_container)
                scripts_state = (scripts_state[0], True)

    async def __call_script(
        self,
        script: Bootscript,
        di_container: DiContainer
    ) -> None:
        dependencies: dict[str, Provider] = collect_dependencies_for_acceptor(
            acceptor_callable=script.fn,
            container=di_container,
            acceptor_module=None
        )

        if inspect.iscoroutinefunction(script.fn):
            await script.fn(**dependencies)
        else:
            script.fn(**dependencies)  # type: ignore



from .configs import LogConfig
from .handlers import LogHandler
from .log import Log
from .utils import LogUtils, configure_log

__all__ = [
    "Log",
    "LogConfig",
    "LogHandler",
    "LogUtils",
    "configure_log",
]



import inspect
import typing
from types import NoneType, UnionType

from orwynn.utils.url import UrlVars
from orwynn.websocket.websocket import Websocket

from .handlers import WebsocketHandler

# Types allowed to be set at handler's function argument annotations
_AllowedTypesUnion = str | int | float
_AllowedTypes: tuple[type] = typing.get_args(_AllowedTypesUnion)
_Annotation = typing.TypeVar("_Annotation")

# Allowed for handler kwargs
HandlerKwargs = dict[str, _AllowedTypesUnion]


def get_handler_kwargs(
    handler: WebsocketHandler,
    url_vars: UrlVars
) -> HandlerKwargs:
    """
    Inspects the given handler and the given url vars to select and convert
    kwargs required by the handler.

    Args:
        handler:
            Websocket handler to inspect.
        url_vars:
            Collected url variables to select from.

    Returns:
        Kwargs dictionary.
    """
    kwargs: HandlerKwargs = {}


    for param in inspect.signature(handler.fn).parameters.values():
        _check_if_duplicate(
            kwargs=kwargs,
            param=param
        )

        if param.annotation is Websocket:
            continue
        elif (
            type(param.annotation) is UnionType
            and any([
                a in _AllowedTypes
                    for a in typing.get_args(param.annotation)
            ])
        ):
            _set_union_param(
                kwargs=kwargs,
                param=param,
                handler=handler,
                url_vars=url_vars
            )
        elif (
            type(param.annotation) is type
            and param.annotation in _AllowedTypes
        ):
            _set_regular_param(
                kwargs=kwargs,
                param=param,
                url_vars=url_vars
            )
        else:
            raise TypeError(
                f"unsupported param with name {param.name} annotation:"
                f" {param.annotation}"
            )

    return kwargs


def _set_regular_param(
    *,
    kwargs: HandlerKwargs,
    param: inspect.Parameter,
    url_vars: UrlVars
) -> None:
    try:
        path_var_value: str = url_vars.path_vars[param.name]
    except KeyError as err:
        # Path variables unlike query ones cannot be defaulted, so here we
        # raise an error
        raise KeyError(
            f"no path variable for name: {param.name}"
        ) from err
    else:
        kwargs[param.name] = _convert_var_value(
            path_var_value,
            Annotation=param.annotation
        )


def _set_union_param(
    *,
    kwargs: HandlerKwargs,
    param: inspect.Parameter,
    handler: WebsocketHandler,
    url_vars: UrlVars
) -> None:
    union_args: tuple = param.annotation.__args__
    _check_union_args(
        union_args,
        handler=handler,
        param=param
    )

    final_value: typing.Any
    try:
        # For union types we operate only with query args
        query_var_value: str = url_vars.query_vars[param.name]

        # Choose non-None element of the union
        PositiveAnnotation: type = \
            union_args[0] if union_args[0] is not NoneType else union_args[1]

        final_value = _convert_var_value(
            query_var_value,
            Annotation=PositiveAnnotation
        )
    except KeyError:
        # Query variables can be omitted: in this case we use default
        # values specified in annotation
        #
        # Also, in the default case no convertation is required since it should
        # be a plain python object which came from the code
        final_value = param.default

    kwargs[param.name] = final_value


def _check_union_args(
    union_args: tuple,
    *,
    handler: WebsocketHandler,
    param: inspect.Parameter
) -> None:
    if param.default is inspect._empty:
        raise ValueError(
            f"handler {handler} param {param.name} with union annotation"
            f" {param.annotation} should have default value since"
            " it will be filled from an optional query"
        )
    elif (
        type(param.default) not in _AllowedTypes
        and param.default is not None
    ):
        raise TypeError(
            f"parameter's default {param.default} is neither in allowed types"
            f" {_AllowedTypes} nor is None"
        )
    elif len(union_args) != 2:
        raise ValueError(
            "only strict unions are allowed: length of handler"
            f" {handler} param {param.name}"
            f" annotation {param.annotation} should equal 2"
        )
    elif NoneType not in union_args:
        raise TypeError(
            f"handler {handler} param {param.name} annotation"
            f" {param.annotation} union args {union_args} should "
            " contain None at any position"
        )
    elif (
        union_args[0] not in _AllowedTypes
        and union_args[1] not in _AllowedTypes
    ):
        raise TypeError(
            f"parameter {param.name} have unsupported element in union"
            f" annotation {param.annotation}"
        )


def _check_if_duplicate(
    *,
    kwargs: HandlerKwargs,
    param: inspect.Parameter
) -> None:
    if param.name in kwargs:
        raise ValueError(
            f"variable {param.name} has been already set in handler kwargs"
        )


def _convert_var_value(
    value: str,
    *,
    Annotation: type[_Annotation]
) -> _Annotation:
    # Apply convertation to one of the allowed formats, now it can be done
    # directly with annotation (after the checks made of course), e.g.
    # with int annotation it would be int(VARIABLE)
    return Annotation(value)



import httpx
from fastapi import Response as FastAPIResponse
from fastapi.responses import FileResponse as FastAPI_FileResponse
from fastapi.responses import HTMLResponse as FastAPI_HTMLResponse
from fastapi.responses import JSONResponse as FastAPI_JSONResponse
from starlette.responses import RedirectResponse as Starlette_RedirectResponse

HttpResponse = FastAPIResponse
JsonHttpResponse = FastAPI_JSONResponse
HtmlHttpResponse = FastAPI_HTMLResponse
TestHttpResponse = httpx.Response
FileHttpResponse = FastAPI_FileResponse
RedirectHttpResponse = Starlette_RedirectResponse



from orwynn.mongo.document.testing import SimpleDocument


def test_main(document_1: SimpleDocument, document_2: SimpleDocument):
    assert {item.id for item in SimpleDocument.get()} == {document_1.id,
        document_2.id}


def test_limited(document_1: SimpleDocument, document_2: SimpleDocument):
    assert len(list(SimpleDocument.get(limit=1))) == 1


def test_id_operators(
    document_1: SimpleDocument,
    document_2: SimpleDocument
):
    """
    Should work normally for id MongoDb operators.
    """
    found: list[SimpleDocument] = list(SimpleDocument.get({
        "id": {
            "$in": [
                document_1.id
            ]
        }
    }))

    assert len(found) == 1
    assert found[0].id == document_1.id






class AlreadyInitializedStorageError(Exception):
    """If the context storage has been already initialized."""


class RequestIdAlreadySavedError(Exception):
    """If the request id is attempted to be saved twice."""


class UndefinedStorageError(Exception):
    """Storage is not set for the current context."""



from orwynn.base.errorhandler import ErrorHandler
from orwynn.websocket.middleware.connectionbuiltin import (
    ConnectionBuiltinWebsocketMiddleware,
)

from .context.middleware.builtin import (
    ContextBuiltinWebsocketMiddleware,
)
from .context.middleware.contextbuiltin import (
    RequestContextBuiltinWebsocketMiddleware,
)
from .errorhandler.default import (
    DefaultWebsocketErrorHandler,
)
from .errorhandler.middleware import (
    ErrorHandlerWebsocketMiddleware,
)
from .middleware.builtin import BuiltinWebsocketMiddleware

DEFAULT_WEBSOCKET_EXCEPTION_HANDLERS: set[type[ErrorHandler]] = {
    DefaultWebsocketErrorHandler,
}

BUILTIN_WEBSOCKET_MIDDLEWARE: list[type[BuiltinWebsocketMiddleware]] = [
    # Connection middleware should be first, since the exception handlers will
    # access websocket object
    ConnectionBuiltinWebsocketMiddleware,
    ContextBuiltinWebsocketMiddleware,
    ErrorHandlerWebsocketMiddleware,
    RequestContextBuiltinWebsocketMiddleware
]



from orwynn.base.service.service import Service


class FrameworkService(Service):
    """Acts most like a generic Service, but is pre-initialized in DI.

    Shouldn't be subclassed in user's applications, only in the framework or
    extensions.

    This service shouldn't be added in module.Providers in order to be
    injected - it is always accessible for any Acceptor in the app.
    """



from typing import Any, Callable

from fastapi.routing import APIWebSocketRoute as _NativeWebsocketRoute
from fastapi.routing import get_websocket_app
from starlette.routing import (
    compile_path,
    get_name,
    websocket_session,
)

from orwynn.app.utils import get_dependant


class ApiWebsocketRoute(_NativeWebsocketRoute):
    def __init__(
        self,
        path: str,
        endpoint: Callable[..., Any],
        *,
        name: str | None = None,
        dependency_overrides_provider: Any | None = None,
    ) -> None:
        # Copy of FastAPI code for APIWebSocket route to fix dependants issue
        # for customized websocket middleware
        #
        self.path = path
        self.endpoint = endpoint
        self.name = get_name(endpoint) if name is None else name
        self.path_regex, self.path_format, self.param_convertors = \
            compile_path(path)

        # Here get_dependant is our function. Due to unknown reason we haven't
        # managed to do the replacement via mock (as we did with ApiRouter for
        # example), so here we copied the entire init function body to replace
        # the function directly
        self.dependant = get_dependant(
            path=self.path_format, call=self.endpoint
        )

        self.app = websocket_session(
            get_websocket_app(
                dependant=self.dependant,
                dependency_overrides_provider=dependency_overrides_provider,
            )
        )



import contextlib
import os
from enum import Enum
from typing import Generator

import pytest
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from orwynn import sql
from orwynn.base.module import Module
from orwynn.boot import Boot
from orwynn.utils import validation
from orwynn.utils.crypto import hash_password

from .config import SqlConfig
from .databasekind import SQLDatabaseKind
from .sql import Sql
from .table import Table


class User(Table):
    name: Mapped[str]
    hpassword: Mapped[str]
    tweets: Mapped[list["Tweet"]] = relationship(backref="user")
    likes: Mapped[list["Like"]] = relationship(backref="user")


class Tweet(Table):
    title: Mapped[str]
    text: Mapped[str]
    likes: Mapped[list["Like"]] = relationship(backref="tweet")
    user_id: Mapped[int] = mapped_column(ForeignKey("user._id"), nullable=True)


class Like(Table):
    user_id: Mapped[int] = mapped_column(
        ForeignKey("user._id"),
        nullable=True
    )
    tweet_id: Mapped[int] = mapped_column(
        ForeignKey("tweet._id"),
        nullable=True
    )


@pytest.fixture
def _sqlite() -> Generator:
    service = Sql(SqlConfig(
        database_kind=SQLDatabaseKind.SQLITE,
        database_path=":memory:"
    ))

    service.create_tables()

    yield service

    service.drop_tables()


@pytest.fixture
def _user1() -> User:
    return User(
        name="Tommy",
        hpassword=hash_password("vicecity"),
    )


@pytest.fixture
def _user2() -> User:
    return User(
        name="Lance",
        hpassword=hash_password("vancenotdance")
    )


@pytest.fixture
def _tweet1() -> Tweet:
    return Tweet(
        title="Ice Cream Factory",
        text="Make 50 sells!"
    )


@pytest.fixture
def _like1() -> Like:
    return Like()


@pytest.fixture
def _add_user1(_sqlite: Sql, _user1: User) -> None:
    with _sqlite.session as session:
        session.add(_user1)
        session.commit()


@pytest.fixture
def _add_user2(_sqlite: Sql, _user2: User) -> None:
    with _sqlite.session as session:
        session.add(_user2)
        session.commit()


@pytest.fixture
def _add_tweet1(_sqlite: Sql, _tweet1: Tweet) -> None:
    with _sqlite.session as session:
        session.add(_tweet1)
        session.commit()


@pytest.fixture
def _add_like1(_sqlite: Sql, _like1: Like) -> None:
    with _sqlite.session as session:
        session.add(_like1)
        session.commit()


@pytest.fixture
def _connect_user1a2_tweet1_like1(
    _sqlite: Sql,
    _add_user1,
    _add_user2,
    _add_tweet1,
    _add_like1
) -> None:
    with _sqlite.session as s:
        user1: User = validation.check(s.get(User, 1))
        user2: User = validation.check(s.get(User, 2))
        tweet1: Tweet = validation.check(s.get(Tweet, 1))
        like1: Like = validation.check(s.get(Like, 1))
        user2.likes.append(like1)
        tweet1.likes.append(like1)
        user1.tweets.append(tweet1)
        s.commit()


def test_sqlite_init_memory():
    sql = Sql(SqlConfig(
        database_kind="sqlite",
        database_path=":memory:"
    ))
    sql.create_tables()
    sql.drop_tables()


def test_sqlite_init_relative_path():
    # Delete old data to ensure auto directory creating is working.
    with contextlib.suppress(FileNotFoundError):
        try:
            os.removedirs(os.path.join(
                os.getcwd(), "var/tmp"
            ))
        except OSError: # Directory not empty
            os.remove(os.path.join(
                os.getcwd(), "var/tmp/test.db"
            ))

    sql = Sql(SqlConfig(
        database_kind="sqlite",
        database_path="var/tmp/test.db"
    ))
    sql.create_tables()
    sql.drop_tables()


def test_sqlite_init_absolute_path():
    # Delete old data to ensure auto directory creating is working.
    with contextlib.suppress(FileNotFoundError):
        try:
            os.removedirs("/tmp/orwynn")  # noqa: S108
        except OSError: # Directory not empty
            os.remove("/tmp/orwynn/test.db")  # noqa: S108

    sql = Sql(SqlConfig(
        database_kind="sqlite",
        database_path="/tmp/orwynn/test.db"  # noqa: S108
    ))
    sql.create_tables()
    sql.drop_tables()


def test_postgresql_init():
    sql = Sql(SqlConfig(
        database_kind="postgresql",
        database_name="orwynn_test",
        database_user="postgres",
        database_password="postgres",  # noqa: S106
        database_host="localhost",
        database_port=9005
    ))
    sql.create_tables()
    sql.drop_tables()


def test_create(
    _sqlite: Sql,
    _connect_user1a2_tweet1_like1
):
    with _sqlite.session as s:
        user2: User = validation.check(s.get(User, 2))
        tweet1: Tweet = validation.check(s.get(Tweet, 1))
        like1: Like = validation.check(s.get(Like, 1))

        assert validation.check(s.get(User, 1)).tweets == [tweet1]
        assert tweet1.likes == [like1]
        assert user2.likes == [like1]


def test_enum_field(_sqlite: Sql):
    class Color(Enum):
        RED = 1
        BLUE = 2
        GREEN = 3

    class Item(Table):
        color: Mapped[Color]

    _sqlite.create_tables(Item)
    with _sqlite.session as s:
        s.add(Item(color=Color.RED))
        s.add(Item(color=Color.BLUE))
        s.add(Item(color=Color.GREEN))
        s.commit()

        assert validation.check(s.get(Item, 1)).color == Color.RED
        assert validation.check(s.get(Item, 2)).color == Color.BLUE
        assert validation.check(s.get(Item, 3)).color == Color.GREEN


@pytest.mark.asyncio
async def test_config_poolclass():
    await Boot.create(
        Module(imports=[sql.module]),
        apprc={
            "prod": {
                "Sql": {
                    "database_kind": "sqlite",
                    "database_path": ":memory:",
                    "poolclass": "StaticPool"
                }
            }
        }
    )



from orwynn.base.module import Module

from .config import MongoConfig
from .document import Document
from .mongo import Mongo

module = Module(
    Providers=[Mongo, MongoConfig],
    exports=[Mongo, MongoConfig]
)



class NoScriptsForCallTimeError(Exception):
    """
    When no bootscripts are added for certain call time.
    """


class ScriptsAlreadyCalledError(Exception):
    """
    If some scripts were already called.
    """



from orwynn.utils.fmt import kebabify, pascalify, snakefy


def test_kebabify():
    assert kebabify("wow_hello") == "wow-hello"


def test_snakefy():
    assert snakefy("HelloWorld") == "hello_world"
    assert snakefy("hello_world") == "hello_world"
    assert snakefy("StandardHTTPS") == "standard_https"
    assert snakefy("standardHTTPS") == "standard_https"
    assert snakefy("standardHTTPSToHTTP") == "standard_https_to_http"
    assert snakefy("HTTPSToHTTPGen") == "https_to_http_gen"
    assert snakefy("HelloWorldD") == "hello_world_d"
    assert snakefy("H") == "h"
    assert snakefy("HE") == "he"
    assert snakefy("HE_WO") == "he_wo"
    assert snakefy("_hello_worldAgain_") == "_hello_world_again_"


def test_pascalify():
    assert pascalify("hello_world") == "HelloWorld"
    assert pascalify("HelloWorld") == "HelloWorld"
    assert pascalify("hello_WORLD_AGAIN") == "HelloWORLDAGAIN"
    assert pascalify("_helloworld_again_") == "_HelloworldAgain_"
    assert pascalify("h") == "H"
    assert pascalify("h_e") == "HE"
    assert pascalify("h_e_world") == "HEWorld"
    assert pascalify("____hello_world____") == "____HelloWorld____"
    assert pascalify("____hello123_w23orld____") == "____Hello123W23orld____"
    assert pascalify("____56hello1_w23orld____") == "____56hello1W23orld____"



import contextlib
import json

from starlette.concurrency import iterate_in_threadpool
from starlette.responses import StreamingResponse

from orwynn.base.error import MalfunctionError
from orwynn.http.requests import HttpRequest
from orwynn.http.responses import HttpResponse
from orwynn.log import Log
from orwynn.utils import validation


class HttpLogger:
    """Logs HTTP requests and responses."""
    async def log_request(
        self,
        request: HttpRequest,
        request_id: str
    ) -> str:
        """Assigns special id to request and logs it.

        Args:
            request:
                Request to be logged.
            request_id:
                ID of the request.

        Returns:
            Request assigned ID.
        """
        validation.validate(request, HttpRequest)
        validation.validate(request_id, str)

        plain_message: str = \
            f"request {request.method.upper()} {request.url.path}" \
            f"{request.url.query}"

        json_: dict | None = None
        with contextlib.suppress(json.JSONDecodeError):
            # FIXME: hangs here, using request.body hangs too, see:
            #   https://github.com/encode/starlette/issues/847
            #
            # Seems like for now no request json will be collected until
            # mentioned issue is resolved.
            json_ = None
        if not json_:
            json_ = None

        extra: dict = {
            "http": {
                "request": {
                    "id": request_id,
                    # Get full URL
                    "url": request.url._url,
                    "headers": dict(request.headers),
                    "json": json_
                }
            }
        }

        # Note that here and on response field "request_id" also duplicated
        # at extra.request_id by logger itself. It's ok and shouldn't be
        # removed for all logs compliance.
        Log.bind(**extra).info(
            plain_message
        )

        return request_id

    async def log_response(
        self,
        response: HttpResponse,
        *,
        request: HttpRequest,
        request_id: str
    ) -> None:
        """Logs a response linking it to the according request.
        """
        # Fetch response body: for StreamingResponse it should apply special
        # logic.
        response_body: str
        if isinstance(response, StreamingResponse):
            try:
                response_body: str = await self.__get_response_body(response)
            except (UnicodeDecodeError, IndexError):
                response_body = ""
        else:
            # Interestingly, even returned explicitly JSONResponse handled here
            # as StreamingResponse, so maybe there is not need to handle
            # "else:" branch here.
            raise MalfunctionError(
                f"unexpected branch entering for response {response}"
            )
            # And this is an alternative handling, consider enabling it if you
            # get the MalfunctionError:

        plain_message: str = \
            f"response {response.status_code}" \
            f" {request.url.path}{request.url.query}:" \
            f" {response_body}"
        json_: dict | None

        content_type: str = response.headers.get("content-type", "")
        if "application/json" in content_type:
            json_ = validation.apply(
                json.loads(
                    response_body
                ),
                dict
            )
            if not json_:
                json_ = None
        else:
            json_ = None

        extra: dict = {
            "http": {
                "response": {
                    "status_code": response.status_code,
                    "request_id": request_id,
                    "media_type": response.media_type,
                    "headers": dict(response.headers),
                    "json": json_
                }
            }
        }

        Log.bind(**extra).info(
            plain_message
        )

    async def __get_response_body(
        self,
        response: StreamingResponse
    ) -> str:
        # Gets response body from StreamingResponse, see:
        #   https://github.com/encode/starlette/issues/874#issuecomment-1027743996
        #
        validation.validate(response, StreamingResponse)

        response_body = [section async for section in response.body_iterator]
        response.body_iterator = iterate_in_threadpool(iter(response_body))
        index0_response_body: bytes = validation.apply(response_body[0], bytes)
        return index0_response_body.decode()



from .constants import (
    BUILTIN_WEBSOCKET_MIDDLEWARE,
    DEFAULT_WEBSOCKET_EXCEPTION_HANDLERS,
)
from .context.id import WebsocketRequestContextId
from .controller.controller import WebsocketController
from .controller.eventhandlermethod import WebsocketEventHandlerMethod
from .errorhandler.middleware import ErrorHandlerWebsocketMiddleware
from .middleware.builtin import BuiltinWebsocketMiddleware
from .middleware.middleware import WebsocketMiddleware
from .middleware.nextcall import WebsocketNextCall
from .routing import handlers as routing_handlers
from .routing.stack import WebsocketStack
from .websocket import Websocket



import copy
from types import NoneType
from typing import Self

from orwynn.base.controller import Controller
from orwynn.base.middleware import Middleware as MiddlewareClass
from orwynn.base.module.errors import (
    CircularDependencyError,
    EmptyRouteError,
    FrameworkServiceModuleReferenceError,
)
from orwynn.base.service.framework import FrameworkService
from orwynn.utils import validation
from orwynn.utils.validation import validate, validate_route


class Module:
    """Provides metadata to organize the application structure.

    Attributes:
        route (optional):
            All controllers defined under this module will
            operate under this route. By default the module has no route and
            hence cannot accept any controllers and middlewares.
        Providers (optional):
            List of Providers to be initialized and shared at least across this
            module.
        Controllers (optional):
            List of Controllers to be initialized for this module.
        Middleware (optional):
            List of Middleware classes applied to all module' controllers.
        imports (optional):
            List of imported modules to use their exported providers in this
            module.
        exports (optional):
            Sublist of providers that are provided by this module for other
            modules importing this module. It cannot contain provider not
            referenced in `providers` field.

    Usage:
    ```py
    # medieval_module.py

    medieval_module = Module(
        Providers=[KingService, KnightService, KingConfig],
        Controllers=[KingController, KnightController],
        Middleware=[BlessingMiddleware],
        imports=[castle_module],
        exports=[KingService, KingConfig]
    )
    ```
    """
    def __init__(
        self,
        route: str | None = None,
        *,
        Providers: list[type] | None = None,
        Controllers: list[type[Controller]] | None = None,
        Middleware: list[type[MiddlewareClass]] | None = None,
        imports: list["Module"] | None = None,
        exports: list[type] | None = None
    ) -> None:
        super().__init__()
        validate(route, [str, NoneType])
        validate(Providers, [list, NoneType])
        validate(Controllers, [list, NoneType])
        validate(Middleware, [list, NoneType])
        validate(imports, [list, NoneType])
        validate(exports, [list, NoneType])

        self.__route: str | None
        if route is None:
            self.__route = route

            # Module without route cannot accept Controller and Middleware
            if Controllers is not None:
                raise ValueError(
                    f"module {self} has not route and cannot accept"
                    " controllers"
                )
            elif Middleware is not None:
                raise ValueError(
                    f"module {self} has not route and cannot accept"
                    " middleware"
                )
        else:
            self.__route = self.__parse_route(route)

        self._Providers: list[type] = self._parse_providers(
            Providers
        )
        self._Controllers: list[type[Controller]] = self._parse_controllers(
            Controllers
        )
        self._Middleware: list[type[MiddlewareClass]] = \
            self._parse_middleware(Middleware)
        self._imports: list["Module"] = self._parse_imports(imports)

        # Validate and assign exports
        if not exports:
            exports = []
        self._exports: list[type] = exports

    def __repr__(self) -> str:
        return "<{} \"{}\" at {}>".format(
            self.__class__.__name__,
            self.route or "no-route",
            hex(id(self))
        )

    @property
    def route(self) -> str | None:
        return self.__route

    @property
    def Providers(self) -> list[type]:
        return copy.copy(self._Providers)

    @property
    def Controllers(self) -> list[type[Controller]]:
        return copy.copy(self._Controllers)

    @property
    def Middleware(self) -> list[type[MiddlewareClass]]:
        return copy.copy(self._Middleware)

    @property
    def imports(self) -> list["Module"]:
        return copy.copy(self._imports)

    @property
    def exports(self) -> list[type]:
        return copy.copy(self._exports)

    @staticmethod
    def __parse_route(route: str) -> str:
        if route:
            validate_route(route)
        else:
            raise EmptyRouteError()

        return route

    def _fw_add_provider_or_skip(self, P: type) -> None:
        if P not in self._Providers:
            self._Providers.append(P)

    def _fw_add_controller_or_skip(self, C: type[Controller]) -> None:
        validate(C, Controller)
        if C not in self._Controllers:
            self._Controllers.append(C)

    def _fw_add_middleware_or_skip(self, Mw: type[MiddlewareClass]) -> None:
        validate(Mw, MiddlewareClass)
        if Mw not in self._Middleware:
            self._Middleware.append(Mw)

    def _fw_add_imports(self, *modules: Self) -> None:
        """Adds imported modules.

        Generally used to add global imports.
        """
        for module in modules:
            validation.validate(module, Module)
            if module in self._imports:
                raise CircularDependencyError(
                    f"module {module} has been already added to containing"
                    f" module {self}"
                )
            self._imports.append(module)

    def _parse_providers(
        self,
        Providers: list[type] | None
    ) -> list[type]:
        res: list[type]

        if Providers:
            for P in Providers:
                if issubclass(P, FrameworkService):
                    raise FrameworkServiceModuleReferenceError(
                        f"a framework service {P} cannot be referenced in"
                        f" the module {self}"
                    )
            res = Providers
        else:
            res = []

        return res

    @staticmethod
    def _parse_controllers(
        Controllers: list[type[Controller]] | None
    ) -> list[type[Controller]]:
        res: list[type[Controller]]

        if Controllers:
            for Controller_ in Controllers:
                validate(Controller_, Controller)
                # Rest validation done by controller itself
            res = Controllers
        else:
            res = []

        return res

    @staticmethod
    def _parse_middleware(
        Middleware_: list[type[MiddlewareClass]] | None
    ) -> list[type[MiddlewareClass]]:
        res: list[type[MiddlewareClass]]

        if Middleware_:
            for M in Middleware_:
                validate(M, MiddlewareClass)
            res = Middleware_
        else:
            res = []

        return res

    @staticmethod
    def _parse_imports(
        imports: list["Module"] | None
    ) -> list["Module"]:
        res: list["Module"]

        if imports:
            for import_ in imports:
                validate(import_, Module)
            res = imports
        else:
            res = []

        return res



import os
from typing import ClassVar

from orwynn.base.error.errors import MalfunctionError
from orwynn.helpers.web import GenericRequest, GenericResponse
from orwynn.log import LogUtils
from orwynn.proxy.boot import BootProxy
from orwynn.utils import validation
from orwynn.utils.scheme import Scheme


class ErrorHandler:
    """
    Handles outcoming errors from the application.
    Method handle(...) should be redefined in subclass in order to work.

    Attributes:
        E:
            Exception or a list of handled Exceptions.
        PROTOCOL:
            Protocol the handler works with.
        IS_ERROR_CATCH_LOGGED:
            Whether the handled errors should be automatically logged by
            Log.catch. Defaults to True.
    """
    E: ClassVar[type[Exception] | None] = None
    PROTOCOL: Scheme = Scheme.HTTP
    IS_ERROR_CATCH_LOGGED: bool = True

    def __init__(self) -> None:
        if self.E is None:
            raise TypeError(
                f"{self.__class__} error class is not set"
            )
        else:
            validation.validate(self.E, Exception)

        validation.validate(self.PROTOCOL, Scheme)

    @classmethod
    def get_handled_exception_class(cls) -> type[Exception]:
        if cls.E is None:
            raise MalfunctionError(
                f"error handler {cls} handled exception shouldn't be None"
            )
        return cls.E

    @property
    def HandledException(self) -> type[Exception]:
        return self.__class__.get_handled_exception_class()

    def _fw_handle_wrapper(
        self,
        request: GenericRequest,
        error: Exception
    ) -> GenericResponse:
        """
        Actual handler called on error occuring.

        Inside it should always propagate the control to self.handle.
        """
        if (
            self.IS_ERROR_CATCH_LOGGED
            # check without an AppMode importing due to circular issues
            and BootProxy.ie().mode.value != "test"
            and not os.getenv(
                "ORWYNN_IS_CATCH_LOGGING_ENABLED_IN_TESTS",
                False
            )
        ):
            LogUtils.catch_error(error)
        return self.handle(request, error)

    def handle(
        self,
        request: GenericRequest,
        error: Exception
    ) -> GenericResponse:
        raise NotImplementedError()



from orwynn.app.app import App
from orwynn.base.controller.controller import Controller
from orwynn.base.errorhandler.errorhandler import ErrorHandler
from orwynn.base.middleware import GlobalMiddlewareSetup, Middleware
from orwynn.base.module.module import Module
from orwynn.base.worker.worker import Worker
from orwynn.di.collecting.modulecollector import ModuleCollector
from orwynn.di.collecting.providerdependencies.collect import (
    collect_provider_dependencies,
)
from orwynn.di.container import DiContainer
from orwynn.di.init.acceptors import init_other_acceptors
from orwynn.di.init.providers import init_providers
from orwynn.di.object import DiObject
from orwynn.http.log.configs import LogHttpMiddlewareConfig
from orwynn.log import LogConfig
from orwynn.utils import validation


class Di(Worker):
    """Resolves Dependency-injection related tasks for an application.

    Only objects in categories Providers and Acceptors take part in DI. For now
    only objects from BUILTIN_PROVIDERS and BUILTIN_ACCEPTORS are considered.
    In future it might be possible to add custom providers or acceptors (maybe
    using class decorators) in addition to builtin ones.

    Stages of DI:
    - Build a dependency tree, collecting all objects and their requested
        injections
    - Make an order of initialization for dependencies
    - Initialize dependencies according to the order for each making
        appropriate injections

    How high in order dependency will be placed depends on:
    - Dependency's priority, see BUILTIN_PROVIDERS
    - Amount of other dependencies and their priorities:
    ```
    dep1 * dep1_priority + dep2 * dep2_priority + ... + dep_n * dep_n_priority
    ```
    - Amount of imports of other modules at the module, where considered
        dependency resides

    Attributes:
        root_module:
            Root module of the app.
        global_modules (optional):
            List of modules to be applied globally.
        global_middleware (optional):
            Middleware setup map to be applied globally.
    """
    def __init__(
        self,
        root_module: Module,
        *,
        global_modules: list[Module] | None = None,
        global_middleware: GlobalMiddlewareSetup | None = None
    ) -> None:
        super().__init__()
        validation.validate(root_module, Module)

        if global_modules is None:
            global_modules = []
        validation.validate_each(
            global_modules, Module, expected_sequence_type=list
        )

        if global_middleware is None:
            global_middleware = {}
        validation.validate_dict(
            global_middleware, (Middleware, list)
        )

        # Add framework services
        root_module._fw_add_provider_or_skip(App)
        # Log configs is always added to configure logging, they all should
        # be able to be built from an empty apprc (all values set to default).
        root_module._fw_add_provider_or_skip(LogConfig)
        root_module._fw_add_provider_or_skip(LogHttpMiddlewareConfig)
        ##

        # So here we have generally a two stages of DI:
        #   1. Collecting (component "di/collecting")
        #   2. Initializing (component "di/init")

        self.modules: list[Module] = ModuleCollector(
            root_module,
            global_modules=global_modules
        ).collected_modules

        self.__container: DiContainer = init_providers(
            collect_provider_dependencies(self.modules)
        )
        init_other_acceptors(self.__container, self.modules, global_middleware)

    @property
    def _fw_container(self) -> DiContainer:
        return self.__container

    @property
    def app_service(self) -> App:
        return validation.apply(self.find("App"), App)

    @property
    def controllers(self) -> list[Controller]:
        """Fetches all controllers from container.

        Returns:
            All controllers fetched.
        """
        return self.__container.controllers

    @property
    def all_middleware(self) -> list[Middleware]:
        """Fetches all middleware from container.

        Returns:
            All middleware fetched.
        """
        return self.__container.all_middleware

    @property
    def exception_handlers(self) -> list[ErrorHandler]:
        return self.__container.exception_handlers

    def find(self, key: str) -> DiObject:
        """Returns DI object by its key.

        Note that searching is made using PascalCased keys, but actual object
        returned is an initialized instance of searched class.

        Args:
            key:
                String value to search with.

        Returns:
            A DIObject found.

        Raises:
            MissingDIObjectError:
                DIObject with given key is not found.
        """
        return self.__container.find(key)



import contextlib
from orwynn.utils.klass import Static
import sys
from typing import TYPE_CHECKING, Any
import typing_extensions
from orwynn.log import Log
from orwynn.log.configs import LogConfig
from orwynn.log.handlers import LogHandler
from orwynn.proxy.boot import BootProxy

if TYPE_CHECKING:
    from orwynn.app import AppMode


class LogUtils(Static):
    catchr = Log.catch(reraise=True)
    """
    Shortcut for Log.catch(reraise=True).

    Does not accept additional arguments.
    """

    @staticmethod
    def catch_error(error: Exception):
        """Reraise error and catch it by logger."""
        # TODO(ryzhovalex):
        #   Find way to remove this last function from exception stack to not
        #   be displayed

        # try-except block for correct linting even if error isn't actually
        # passed up the stack
        with contextlib.suppress(Exception), Log.catch(reraise=False):
            raise error

    @staticmethod
    def configure_log(config: LogConfig, *, app_mode_prod: "AppMode") -> None:
        if config.handlers:
            for handler in config.handlers:
                LogUtils._add_handler(handler, app_mode_prod)

    @staticmethod
    def _add_handler(handler: LogHandler, app_mode_prod: "AppMode") -> int:
        """Adds a new handler.

        Args:
            handler:
                Log handler object to be added.

        Returns:
            Added handler number.
        """
        sink: Any
        if isinstance(handler.sink, str):
            # For apprc-based configuration parse some string literals into
            # actual objects.
            match handler.sink:
                case "$stdout":
                    sink = sys.stdout
                case "$stderr":
                    sink = sys.stderr
                case _:
                    # For all other string cases consider the sink as a file
                    # path
                    sink = handler.sink
        else:
            # the sink is given directly to the method as a python object,
            # i.e. add it as it is
            sink = handler.sink

        if BootProxy.ie().mode == app_mode_prod:
            if handler.level is None:
                handler.level = "INFO"
            if handler.serialize is None:
                handler.serialize = True
        else:
            if handler.level is None:
                handler.level = "DEBUG"
            if handler.serialize is None:
                handler.serialize = False

        handler_kwargs: dict = handler.kwargs or {}
        kwargs: dict = dict(
            level=handler.level,
            format=handler.format,
            serialize=handler.serialize,
            **handler_kwargs
        )
        if isinstance(handler.sink, str):
            kwargs["rotation"] = handler.rotation  # type: ignore

        return Log.add(
            sink,
            **kwargs
        )


@typing_extensions.deprecated("use LogUtils")
def configure_log(config: LogConfig, *, app_mode_prod: "AppMode") -> None:
    LogUtils.configure_log(config, app_mode_prod=app_mode_prod)



"""Operations with strings.

Typically here collected functions which return string.
"""
import re
from typing import Any


def format_chain(lst: list[Any]) -> str:
    return " -> ".join([str(x) for x in lst])


def kebabify(name: str) -> str:
    """Converts name to kebab-case.

    Args:
        name:
            Name to convert.

    Returns:
        Name converted.
    """
    # All underscores are simply replaced
    return name.lower().replace("_", "-")


def snakefy(name: str) -> str:
    """Converts name to snake_case.

    Note that this is not reversible using camelfy().

    Args:
        name:
            Name to convert.

    Returns:
        Name converted.
    """
    # Reference: https://stackoverflow.com/a/1176023/14748231
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    name = re.sub("__([A-Z])", r"_\1", name)
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
    return name.lower()


def pascalify(name: str) -> str:
    """Converts name to PascalCase.

    Args:
        name:
            Name to convert.

    Returns:
        Name converted.
    """
    result: str = ""
    is_previous_underscore: bool = True

    for i, char in enumerate(name):
        if is_previous_underscore:
            if char == "_":
                result += char
            else:
                result += char.upper()
        else:
            if char != "_":
                result += char
            else:
                try:
                    next_char: str = name[i + 1]
                except IndexError:
                    result += char
                else:
                    if next_char == "_":
                        result += char
        is_previous_underscore = char == "_"

    return result



import re


def match_routes(
    *,
    abstract_route: str,
    real_route: str
) -> re.Match | None:
    """
    Returns match object of comparing real route against an abstract one.
    """
    regex_ready_abstract_route: str = re.sub(
        r"\\{\w+\\}",
        r"(\\w+)",
        re.escape(abstract_route)
    )
    return re.fullmatch(
        re.compile(regex_ready_abstract_route),
        real_route
    )



import inspect

import pytest

from orwynn.base.config import Config
from orwynn.base.module import Module
from orwynn.di.collecting.modulecollector import ModuleCollector
from orwynn.di.collecting.providerdependencies.collect import (
    collect_provider_dependencies,
)
from orwynn.di.collecting.providerdependencies.map import (
    ProviderDependenciesMap,
)
from orwynn.di.isprovider import is_provider
from orwynn.di.provider import Provider
from orwynn.http.log.configs import LogHttpMiddlewareConfig
from orwynn.log import LogConfig
from tests.std.assertion import Assertion


@pytest.fixture
def std_provider_dependencies_map(
    std_modules: list[Module]
) -> ProviderDependenciesMap:
    return collect_provider_dependencies(std_modules)


def test_std(std_struct: Module):
    # TODO(ryzhovalex):
    #   Here we out of DI adding core providers, which is a duplicate
    #   operation. Need to consider better solutions.
    std_struct._fw_add_provider_or_skip(LogConfig)
    std_struct._fw_add_provider_or_skip(LogHttpMiddlewareConfig)

    metamap: ProviderDependenciesMap = collect_provider_dependencies(
        ModuleCollector(std_struct).collected_modules
    )

    # Order doesn't matter
    assert set(metamap.Providers) == set(Assertion.COLLECTED_PROVIDERS)

    for P, dependencies in metamap.mapped_items:
        assertion_dependencies: list[type[Provider]] = []
        for inspect_parameter in inspect.signature(P).parameters.values():
            # Skip config's parseable parameters
            if (
                issubclass(P, Config)
                and (
                    not inspect.isclass(inspect_parameter.annotation)
                    or not is_provider(inspect_parameter.annotation)
                )
            ):
                continue
            if inspect_parameter.name in ["args", "kwargs"]:
                continue
            assertion_dependencies.append(inspect_parameter.annotation)
        assert dependencies == assertion_dependencies



from orwynn.mongo.document.errors import InvalidOperatorError
from orwynn.mongo.document.testing import SimpleDocument
from orwynn.utils import validation


def test_min(document_1: SimpleDocument):
    document_1 = document_1.update(
        operators={
            "$min": {
                # less than specified
                "price": 0.5
            }
        }
    )

    assert document_1.price == 0.5


def test_min_fail(document_1: SimpleDocument):
    document_1 = document_1.update(
        operators={
            "$min": {
                # more than specified
                "price": 1.5
            }
        }
    )

    # the same amount should remain
    assert document_1.price == 1.2


def test_set_in_operators(document_1: SimpleDocument):
    """
    $set shouldn't appear in `operators` argument.
    """
    validation.expect(
        document_1.update,
        InvalidOperatorError,
        operators={
            "$set": {
                "price": 5
            }
        }
    )


def test_inc_in_operators(document_1: SimpleDocument):
    """
    $inc shouldn't appear in `operators` argument.
    """
    validation.expect(
        document_1.update,
        InvalidOperatorError,
        operators={
            "$inc": {
                "price": 5
            }
        }
    )



from orwynn.base.config import Config
from orwynn.base.module import Module
from orwynn.base.module.errors import CircularDependencyError
from orwynn.di.availability import check_availability
from orwynn.di.collecting.errors import (
    ProviderAlreadyInitializedForMapError,
)
from orwynn.di.collecting.helpers import (
    get_parameters_for_provider,
)
from orwynn.di.collecting.providerdependencies.map import (
    ProviderDependenciesMap,
)
from orwynn.di.errors import NotProviderError
from orwynn.di.isprovider import is_provider
from orwynn.di.provider import Provider
from orwynn.utils.fmt import format_chain


def collect_provider_dependencies(
    modules: list[Module]
) -> ProviderDependenciesMap:
    """Collects providers and their dependencies from given modules.

    Args:
        modules:
            List of modules to collect providers from.

    Returns:
        Special structure maps providers and their dependencies.
    """
    metamap: ProviderDependenciesMap = ProviderDependenciesMap()

    # Traverse all parameters of all providers in all modules to add them in
    # united structure
    for module in modules:
        for P in module.Providers:
            # Chain is cleared for every new provider iterated
            _traverse(P, metamap, [], module)

    return metamap


def _traverse(
    P: type[Provider],
    metamap: ProviderDependenciesMap,
    chain: list[type[Provider]],
    target_module: Module | None
):
    # Recursively traverses over Provider parameters, memorizing chain to
    # find circular dependencies and saving results to metamap.  Target module
    # required since we want to check availability between providers.

    if P in chain:
        raise CircularDependencyError(
            # Failed provider is added second time to the chain for
            # error descriptiveness
            f"provider {P} occured twice in dependency chain"
            f" {format_chain(chain + [P])}"
        )
    chain.append(P)

    try:
        metamap.init_provider(P)
    except ProviderAlreadyInitializedForMapError:
        pass
    else:
        for parameter in get_parameters_for_provider(P):
            if not is_provider(parameter.DependencyProvider):
                # Config's fields are skipped since they request various other
                # than provider objects and will be initialized in a bit
                # different way
                if issubclass(P, Config):
                    continue
                raise NotProviderError(
                    FailedClass=parameter.DependencyProvider
                )

            nested_module: Module | None = check_availability(
                P,
                parameter.DependencyProvider,
                target_module
            )
            metamap.add_dependency(
                dependency=parameter.DependencyProvider,
                P=P
            )
            # And continue traversing, but deeper for fetched dependency
            _traverse(
                parameter.DependencyProvider,
                metamap,
                chain,
                nested_module
            )

    # Pop blocking element. For this concept see collect_modules._traverse
    # function
    chain.pop()



import pytest

from orwynn.base.module import Module
from orwynn.boot.boot import Boot
from orwynn.di.di import Di
from orwynn.http import HttpMiddleware
from orwynn.utils import validation
from orwynn.websocket import WebsocketMiddleware


@pytest.mark.asyncio
async def test_http():
    """
    Should be able to add custom middleware globally.
    """
    class Mw(HttpMiddleware):
        def __init__(self, covered_routes: list[str]) -> None:
            self.covered_routes_ = covered_routes
            super().__init__(covered_routes)

    await Boot.create(
        Module(),
        global_middleware={
            Mw: ["/hello"]
        }
    )

    mw: Mw = validation.apply(Di.ie().find("Mw"), Mw)
    assert mw.covered_routes_ == ["/hello"]


@pytest.mark.asyncio
async def test_websocket():
    """
    Should be able to add custom middleware globally.
    """
    class Mw(WebsocketMiddleware):
        def __init__(self, covered_routes: list[str]) -> None:
            self.covered_routes_ = covered_routes
            super().__init__(covered_routes)

    await Boot.create(
        Module(),
        global_middleware={
            Mw: ["/hello"]
        }
    )

    mw: Mw = validation.apply(Di.ie().find("Mw"), Mw)
    assert mw.covered_routes_ == ["/hello"]



from copy import deepcopy
from typing import Any

import dictdiffer

from orwynn.utils import validation
from orwynn.utils.mp.dictpp import dictpp


def find(location: str, mp: dict) -> Any:
    """Finds field by location.

    Args:
        location:
            String representing location to search from in format
            "key1.key2.key3".
        mp:
            Dict to search in.

    Returns:
        Field found.
    """
    validation.validate(location, str)
    validation.validate(mp, dict)

    return dictpp(mp)[location]


def patch(
    to_be_patched: dict,
    source: dict,
    should_deepcopy: bool = True
) -> dict:
    """Patch one dictionary by another.

    Args:
        to_be_patched:
            Dict to be patched.
        source:
            Source dict to use for patching.
        should_deepcopy (optional):
            Whether deepcopy() should be performed on patched dict. Defaults
            to True. If False is passed, it is convenient to not accept
            returned dict since it is the same as passed one to patch.

    Returns:
        Patched dictionary.
    """
    validation.validate(to_be_patched, dict)
    validation.validate(source, dict)

    patched: dictpp

    # Cast to dictpp to operation with locations. At return it's important
    # to cast it back to normal dict.
    patched = dictpp(
        deepcopy(to_be_patched) if should_deepcopy else to_be_patched
    )

    diff = dictdiffer.diff(to_be_patched, source)

    event_name: str
    location: str
    change: Any

    for event in diff:
        # Unpack event tuple
        event_name, location, change = event
        # How dictdiffer event_name, location and change would like for a
        # reference:
        #   change BurgerShot.menu.cola (1.5, 1.8)
        #   add BurgerShot.menu [('pizza', 4.1), ('fried_chicken', 3.5)]

        # Consider only adding and changing events
        if event_name == "add":
            final_location: str
            for addition in change:
                # Calculate final location in patched dict
                final_location = ".".join(
                    [location, addition[0]]
                ) if location else addition[0]
                patched[final_location] = addition[1]

        elif event_name == "change":
            if location == "":
                raise ValueError(
                    "on change events location shouldn't be empty"
                )
            patched[location] = change[1]

    # Cast back to normal dictionary
    return dict(patched)



class InvalidAddressError(Exception):
    def __init__(
        self,
        *,
        address: str
    ):
        message: str = f"address <{address}> is invalid"
        super().__init__(message)



from datetime import datetime, timedelta, timezone

from orwynn.utils import validation
from orwynn.utils.types import Delta, Timestamp


def get_utc_timestamp() -> Timestamp:
    return datetime.now(timezone.utc).timestamp()


def get_delta_timestamp(delta: Delta) -> Timestamp:
    """Calculates delta timestamp from current moment adding given delta in
    seconds.
    """
    validation.validate(delta, Delta)
    return (
        datetime.now(timezone.utc) + timedelta(seconds=delta)
    ).timestamp()



from orwynn.base.model import Model

from .response import EndpointResponse


class Endpoint(Model):
    """Specification of endpoint.

    Attributes:
        method:
            Method to be used for this endpoint.
        default_status_code (optional):
            Status code to be returned by default from the endpoint. Defaults
            to 200.
        summary (optional):
            Short description of the endpoint. Note that full description is
            taken from the endpoint's docstring. Defaults to pattern summary
            depending on given envpoint's method, e.g. for method POST you
            will get summary "Create {your_model_name}".
        tags (optional):
            List of tags to apply for this endpoint. Empty by default
        is_deprecated (optional):
            Whether this route is deprecated. Defaults to False.
        responses (optional):
            Map with responses representations by their status codes. Defaults
            to built in framework basic responses.
    """
    method: str
    default_status_code: int = 200
    summary: str | None = None
    tags: list[str] | None = None
    is_deprecated: bool = False
    responses: list[EndpointResponse] | None = None



from orwynn.context.manager import context_manager
from orwynn.http import HttpNextCall, HttpRequest, HttpResponse
from orwynn.websocket.middleware.builtin import (
    BuiltinWebsocketMiddleware,
)


class ContextBuiltinWebsocketMiddleware(BuiltinWebsocketMiddleware):
    """Creates a shared context storage active within applied request-response
    cycle.
    """
    async def process(
        self, request: HttpRequest, call_next: HttpNextCall
    ) -> HttpResponse:
        with context_manager():
            return await call_next(request)



from pathlib import Path
from typing import Any, Self

from orwynn.app import AppMode
from orwynn.apprc.apprc import AppRc
from orwynn.base.config import Config
from orwynn.indication.indication import Indication
from orwynn.proxy.boot import BootProxy


class BootConfig(Config):
    """Contains information about boot.

    Attributes:
        mode:
            Boot mode.
        root_dir:
            Root directory of the boot.
    """
    mode: AppMode
    root_dir: Path
    api_indication: Indication
    app_rc: AppRc

    @classmethod
    def load(cls, *, extra: dict[str, Any]) -> Self:
        proxy: BootProxy = BootProxy.ie()
        return cls(
            **proxy.boot_config_data, **extra
        )



import contextlib
import os
from pathlib import Path
from types import NoneType

from orwynn.app import AppMode
from orwynn.apprc.apprc import AppRc
from orwynn.utils import validation
from orwynn.utils.mp import patch as mp_patch
from orwynn.utils.yml import load_yml

from .constants import APP_RC_MODE_NESTING
from .errors import AppRcSearchError


def parse_apprc(
    root_dir: Path,
    mode: AppMode,
    direct_apprc: AppRc | None
) -> AppRc:
    """Parses apprc dictionary.

    Args:
        root_dir:
            Root dir of app.
        mode:
            Boot mode of app.
        direct_apprc:
            AppRc raw configuration passed directly instead of path to it via
            environ. This arg is not optional and should accept None directly,
            if such apprc is not present to enable environ searching way.
    """
    validation.validate(root_dir, Path)
    validation.validate(mode, AppMode)
    validation.validate(direct_apprc, [AppRc, NoneType])

    # All required for this enabled mode data goes here
    final_apprc: AppRc = {}

    if not direct_apprc:
        rc_path_env: str = os.getenv(
            "ORWYNN_APPRC_PATH",
            ""
        )
        should_raise_search_error: bool

        rc_path: Path
        if not rc_path_env:
            rc_path = Path(root_dir, "apprc.yml")
            # On default assignment no errors raised if files not found / empty
            should_raise_search_error = False
        else:
            # Env path started from "./" is supported in this case of
            # concatenation since pathlib.Path does smart path joining
            rc_path = Path(root_dir, rc_path_env)
            should_raise_search_error = True

        if Path(rc_path).exists():
            # Here goes all data contained in yaml config
            apprc: AppRc = load_yml(rc_path)
            __parse_into(final_apprc, apprc, should_raise_search_error, mode)
        elif (
            rc_path_env.startswith("http://")
            or rc_path_env.startswith("https://")
        ):
            raise NotImplementedError("URL sources are not yet implemented")

        elif should_raise_search_error:
            raise AppRcSearchError(
                f"unsupported apprc path {rc_path}"
            )
    else:
        # For direct app rc emptiness check should be always performed, so we
        # pass True
        __parse_into(final_apprc, direct_apprc, True, mode)

    return final_apprc


def __parse_into(
    receiver: dict,
    source: dict,
    should_check_if_source_empty: bool,
    mode: AppMode
) -> None:
    if source == {} and should_check_if_source_empty:
        raise ValueError("parsed apprc source is empty")

    # Check if apprc contains any unsupported top-level keys
    for k in receiver:
        supported_top_level_keys: list[str] = [
            x.value for x in AppMode
        ]
        if k not in supported_top_level_keys:
            raise ValueError(
                f"unsupported top-level key \"{k}\" of apprc config"
            )

    # Load from bottom to top updating previous one with newest one
    mode_nesting_index: int = APP_RC_MODE_NESTING.index(mode)
    for nesting_mode in APP_RC_MODE_NESTING[:mode_nesting_index + 1]:
        # Supress: We don't mind if any top-level key is missing here
        with contextlib.suppress(KeyError):
            mp_patch(
                receiver,
                source[nesting_mode.value],
                should_deepcopy=False
            )



from typing import Callable, ClassVar, Literal

from orwynn.base.controller.controller import Controller
from orwynn.utils import validation
from orwynn.websocket.controller.eventhandlermethod import (
    WebsocketEventHandlerMethod,
)


class WebsocketController(Controller):
    """Operates on websocket protocol.

    Subclasses should define class attribute ROUTE signifies route under
    which all operations is made.

    To define subroute operation, define method started with "on_", rest part
    will be transformed into subroute. Note that snake_case names are
    transformed to kebab-case.

    You can define a special method called "main" - it will be assigned under
    controller's namespace itself (see Example for details).

    Note that final route for each websocket controller method consists of
    three things: MODULE_ROUTE + CONTROLLER_ROUTE + METHOD_ROUTE.

    Class-Attributes:
        ROUTE:
            A subroute of Module's route (where controller attached) which
            controller will answer to. This attribute is required to be
            defined in subclasses explicitly or an error will be raised.
            It is allowed to be "/" to handle Module's root route requests.
        VERSION (optional):
            An API version the controller supports. Defaults to the latest one.

    Example:
    ```python
    from orwynn import WebsocketController

    class MyWSController(WebsocketController):
        ROUTE = "/chat"

        def main(...):  # -> "/chat"
            ...

        def on_connect(...):  # -> "/chat/connect"
            ...

        def on_personal_message(...)  # -> "/chat/personal-message
            ...
    ```
    """
    ROUTE: ClassVar[str | None] = None
    VERSION: ClassVar[int | set[int] | Literal["*"] | None] = None

    @classmethod
    def get_handler_subroutes(cls) -> list[str]:
        """Returns subroutes for each defined handler method."""
        subroutes: list[str] = []
        for k in cls.__dict__:
            if k[:3] == "on_" or k == "main":
                subroutes.append(
                    "/" if k == "main" else k.replace("on_", "/")
                )
        return subroutes

    @property
    def event_handlers(self) -> list[WebsocketEventHandlerMethod]:
        """
        Returns all self bound event handlers started with "on_".

        Note that implicitly inherited methods are not included in search. If
        you want to inherit the parent's functionality for a handler,
        use super() with a redefinition of the method.
        """
        events: list[WebsocketEventHandlerMethod] = []

        for k, v in self.__class__.__dict__.items():
            if k[:3] == "on_" or k == "main":
                try:
                    validation.validate(v, Callable)
                except validation.ValidationError as err:
                    raise ValueError(
                        "every websocket controller object starting from"
                        f" \"on_\" or \"main\" should be Callable, but for"
                        f" name {k} got {v} instead"
                    ) from err

                event_name: str
                event_name = "main" if k == "main" else k[3:]

                if event_name == "":
                    raise ValueError(
                        f"malformed websocket controller method {v} name -"
                        f" {event_name}"
                    )

                events.append(WebsocketEventHandlerMethod(
                    name=event_name,
                    # It's important to use "getattr()" here instead of
                    # "v" since last is obtained from class and is not bound to
                    # an instance.
                    fn=getattr(self, k)
                ))

        if events == []:
            raise ValueError(
                "no event handlers are defined for websocket"
                f" controller {self}"
            )
        return events



import pytest_asyncio

from orwynn.base.module.module import Module
from orwynn.boot.boot import Boot
from orwynn.di.collecting.providerdependencies.map import (
    ProviderDependenciesMap,
)
from orwynn.di.container import DiContainer
from orwynn.di.init.providers import init_providers


@pytest_asyncio.fixture
async def std_di_container(
    std_struct: Module,
    std_provider_dependencies_map: ProviderDependenciesMap
) -> DiContainer:
    await Boot.create(std_struct)
    return init_providers(std_provider_dependencies_map)



import re
from typing import TypeVar

from orwynn.base.config import Config
from orwynn.base.controller.controller import Controller
from orwynn.base.error.errors import MalfunctionError
from orwynn.base.errorhandler.errorhandler import ErrorHandler
from orwynn.base.middleware import Middleware
from orwynn.base.model.model import Model
from orwynn.di.errors import (
    DiObjectAlreadyInitializedInContainerError,
    FinalizedDiContainerError,
    MissingDiObjectError,
)
from orwynn.di.isprovider import is_provider
from orwynn.di.object import DiObject
from orwynn.helpers.constants import SUBCLASSABLES
from orwynn.utils.validation import validate

_InnerObj = TypeVar("_InnerObj")


class DiContainer:
    """Holds data about initialized di objects."""
    def __init__(self) -> None:
        self._data: dict[str, DiObject] = {}

        # Struct running in parallel to main data one to hold references by
        # class for self properties like getting all controllers
        self._objects_by_base_class: dict[type[DiObject], list[DiObject]] = {}

        self._is_provider_populating_finalized: bool = False

    @property
    def items(self) -> list[tuple[str, DiObject]]:
        result: list[tuple[str, DiObject]] = []

        for name, obj in self._data.items():
            result.append((name, obj))

        return result

    @property
    def controllers(self) -> list[Controller]:
        """Fetches all controllers from container.

        Returns:
            All controllers fetched.
        """
        result: list[Controller] = self._find_objects_for_class(Controller)
        return result

    @property
    def all_middleware(self) -> list[Middleware]:
        """Fetches all middleware from container.

        Returns:
            All middleware fetched.
        """
        result: list[Middleware] = self._find_objects_for_class(Middleware)
        return result

    @property
    def exception_handlers(self) -> list[ErrorHandler]:
        result: list[ErrorHandler] = self._find_objects_for_class(
            ErrorHandler
        )
        return result

    def add(self, obj: DiObject) -> None:
        """Adds DI object to the container.

        Args:
            obj:
                DI object to add.

        Raises:
            DIObjectAlreadyInitializedError:
                This DI object already exists in container.
        """
        if self._is_provider_populating_finalized and is_provider(type(obj)):
            raise FinalizedDiContainerError(
                f"cannot add provider {obj}, container is finalized"
            )

        obj_class_name: str = obj.__class__.__name__

        if obj_class_name in self._data:
            raise DiObjectAlreadyInitializedInContainerError(
                f"DI object {obj} already initialized"
            )

        self._data[obj_class_name] = obj

        self._assign_obj_to_base_class(obj)

    def find(self, key: str) -> DiObject:
        """Returns DI object by its key.

        Note that searching is made using PascalCased keys, but actual object
        returned is an initialized instance of searched class.

        Args:
            key:
                String value to search with.

        Returns:
            A DIObject found.

        Raises:
            MissingDIObjectError:
                DIObject with given key is not found.
        """
        validate(key, str)

        try:
            return self._data[key]
        except KeyError as error:
            raise MissingDiObjectError(
                f"di object for key \"{key}\" is not found"
            ) from error

    def find_re(self, pattern: str) -> list[DiObject]:
        """Searches di objects by pattern.

        Args:
            pattern:
                Pattern to apply.

        Returns:
            All di objects with keys matched given pattern.
        """
        validate(pattern, str)

        result: list[DiObject] = []

        for k, v in self._data.items():
            if re.match(pattern, k):
                result.append(v)

        if result == []:
            raise MissingDiObjectError(
                "di objects for pattern \"pattern\" are not found"
            )

        return result

    def finalize_provider_populating(
        self
    ) -> None:
        """Finalizes the container applying logic after provider populating.
        """
        if self._is_provider_populating_finalized:
            raise ValueError("already finalized")

        self._is_provider_populating_finalized = True

    def _assign_obj_to_base_class(self, obj: DiObject) -> None:
        ObjCls: type[DiObject] = type(obj)
        BaseCls: type[DiObject] | None = None

        # Find out object's base class
        for C in SUBCLASSABLES:
            if issubclass(ObjCls, C):
                # Choose more specific base classes for subclasses of Model
                if C is Model and issubclass(ObjCls, Config):
                    BaseCls = Config
                else:
                    BaseCls =  C
        if not BaseCls:
            raise MalfunctionError(
                f"cannot found base class for obj {obj}, this object"
                " shouldn't be appeared in container"
            )

        try:
            self._objects_by_base_class[BaseCls].append(obj)
        except KeyError:
            self._objects_by_base_class[BaseCls] = [obj]

    def _find_objects_for_class(self, C: type[_InnerObj]) -> list[_InnerObj]:
        result: list[_InnerObj] = []
        is_missing: bool = False

        try:
            result = self._objects_by_base_class[C]
        except KeyError:
            is_missing = True
        else:
            if result == []:
                is_missing = True

        if is_missing:
            raise MissingDiObjectError(
                f"cannot find any object for class {C}"
            )
        else:
            return result



from copy import copy

from orwynn.utils import validation

from .errors import UnsupportedVersionError


class ApiVersion:
    """Describes project's API versioning rules.

    Attributes:
        supported (optional):
            Set of version numbers supported. Defaults to only one supported
            version v1.
    """
    def __init__(
        self,
        supported: set[int] | None = None
    ) -> None:
        if supported is None:
            supported = {1}
        validation.validate_each(supported, int, expected_sequence_type=set)
        self.__supported: set[int] = supported

    @property
    def supported(self) -> set[int]:
        return copy(self.__supported)

    @property
    def latest(self) -> int:
        return max(self.__supported)

    def apply_version_to_route(
        self,
        route: str,
        version: int | None = None
    ) -> str:
        """
        Applied a given version to a given route.

        Args:
            route:
                Route to apply version to.
            version (optional):
                Number of version to apply. By default the latest is applied.
        """
        if version is None:
            version = self.latest
        self.check_if_supported(version)

        if "{version}" not in route:
            raise ValueError(
                "no version format block in given route to insert the version"
            )

        return route.replace(
            "{version}",
            str(version)
        )

    def check_if_supported(self, version: int) -> None:
        """Checks if a version is supported.

        Raises:
            UnsupportedVersionError:
                The given version is not supported.
        """
        validation.validate(version, int)
        if version not in self.__supported:
            raise UnsupportedVersionError(
                f"version {version} is not supported"
            )



from orwynn.di.object import DiObject
from orwynn.di.provider import Provider


class MissingDiObjectError(Exception):
    pass


class FinalizedDiContainerError(Exception):
    """If some evil force is trying to add objects to a finalized container.
    """


class DiObjectAlreadyInitializedInContainerError(Exception):
    """DI object has been already initialized.
    """


class NoAnnotationError(Exception):
    """If di object has attribute without annotation."""



class NotProviderError(Exception):
    def __init__(
        self,
        message: str = "",
        FailedClass: type | None = None
    ) -> None:
        if not message and FailedClass:
            message = "{} is not a Provider".format(FailedClass)
        super().__init__(message)


class ProviderAvailabilityError(Exception):
    """If some Provider is not visible for some other Di object."""
    def __init__(
        self,
        message: str = "",
        Provider1: type[Provider] | None = None,
        DiObject: type[DiObject] | None = None
    ) -> None:
        if not message and Provider1 and DiObject:
            message = "provider {} is not available for di object {}".format(
                DiObject, Provider1
            )
        super().__init__(message)



import typing
from typing import Any

from orwynn.base.model.model import Model
from orwynn.mapping.errors import MappingNotLinkedError, UnsetIdMappingError
from orwynn.utils.types import DecoratedCallable


def if_linked(
    fn: DecoratedCallable
) -> DecoratedCallable:
    def inner(self: Mapping, *args, **kwargs):
        if not self.is_linked:
            raise MappingNotLinkedError(
                f"{self} is not linked to database yet"
            )
        return fn(self, *args, **kwargs)
    return typing.cast(DecoratedCallable, inner)


class Mapping(Model):
    """Encapsulates logic of working with database for chosen MODEL.

    Makes a tree of other subclasses which describe interactions with certain
    database kind. End framework user should create their own Mapping
    subclasses and define MODEL and other required class attributes.
    """
    id: str | None = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    def getid(self) -> str:
        if self.id is None:
            raise UnsetIdMappingError(
                explanation="cannot get an id",
                mapping=self
            )
        else:
            return self.id

    def is_linked(self) -> bool:
        """Whether this mapping is linked to actual object in database.
        """
        return self.id is not None

    @classmethod
    def find_all(cls, *args, **kwargs) -> Any:
        raise NotImplementedError()

    @classmethod
    def find_one(cls, *args, **kwargs) -> Any:
        raise NotImplementedError()

    def create(self, *args, **kwargs) -> Any:
        raise NotImplementedError()

    def update(self, *args, **kwargs) -> Any:
        raise NotImplementedError()

    def remove(self, *args, **kwargs) -> Any:
        raise NotImplementedError()
